import gleam/int
import gleam/io
import gleam/list
import gleam/option.{None, Some}
import gleam/order
import gleam/string
import gleam/time/calendar.{Date, TimeOfDay}
import gleam/time/duration
import gleam/time/timestamp.{type Timestamp}
import gleam/uri
import gleam_community/ansi
import spacetraders_api.{type ServerStatus}
import spacetraders_models/account.{type Account}
import spacetraders_models/account_id
import spacetraders_models/agent.{type Agent}
import spacetraders_models/agent_symbol
import spacetraders_models/contract.{type Contract}
import spacetraders_models/contract_id
import spacetraders_models/contract_type
import spacetraders_models/crew_rotation
import spacetraders_models/faction_symbol
import spacetraders_models/ship.{type Ship}
import spacetraders_models/ship_fuel_consumed.{ShipFuelConsumed}
import spacetraders_models/ship_nav_flight_mode
import spacetraders_models/ship_nav_status
import spacetraders_models/ship_role
import spacetraders_models/ship_symbol
import spacetraders_models/trade_symbol
import spacetraders_models/waypoint_symbol
import spacetraders_models/waypoint_type

fn legible_timestamp(timestamp: Timestamp) -> String {
  let #(
    Date(year:, month:, day:),
    TimeOfDay(hours:, minutes:, seconds:, nanoseconds: _),
  ) = timestamp.to_calendar(timestamp, calendar.local_offset())
  int.to_string(hours)
  <> ":"
  <> int.to_string(minutes)
  <> ":"
  <> int.to_string(seconds)
  <> " "
  <> int.to_string(day)
  <> "/"
  <> int.to_string(calendar.month_to_int(month))
  <> "/"
  <> int.to_string(year)
}

fn legible_difference(left: Timestamp, right: Timestamp) -> String {
  let #(value, unit) = timestamp.difference(left, right) |> duration.approximate
  case value == 0 {
    True -> "now"
    False -> {
      let s =
        int.to_string(value)
        <> " "
        <> case unit {
          duration.Year -> "year"
          duration.Month -> "month"
          duration.Week -> "week"
          duration.Day -> "day"
          duration.Hour -> "hour"
          duration.Minute -> "minute"
          duration.Second -> "second"
          duration.Millisecond -> "millisecond"
          duration.Microsecond -> "microsecond"
          duration.Nanosecond -> "nanosecond"
        }
        <> case value {
          1 | -1 -> ""
          _ -> "s"
        }
      case value > 0 {
        True -> "in " <> s
        False -> s <> " ago"
      }
    }
  }
}

pub fn account(account: Account) -> Nil {
  io.println(ansi.bold(ansi.underline(ansi.blue("Account"))))
  io.println(ansi.bold("Id:\t\t") <> account_id.to_string(account.id))
  case account.email {
    Some(email) -> io.println(ansi.bold("Email:\t\t") <> email)
    None -> Nil
  }
  io.println(
    ansi.bold("Created At:\t") <> legible_timestamp(account.created_at),
  )
}

pub fn agent(agent: Agent) -> Nil {
  io.println(ansi.bold(ansi.underline(ansi.magenta("Agent"))))
  io.println(ansi.bold("Symbol:\t\t") <> agent_symbol.to_string(agent.symbol))
  io.println(
    ansi.bold("Headquarters:\t")
    <> waypoint_symbol.to_string(agent.headquarters),
  )
  io.println(ansi.bold("Credits:\t") <> int.to_string(agent.credits))
  io.println(
    ansi.bold("Faction:\t") <> faction_symbol.to_string(agent.starting_faction),
  )
  io.println(ansi.bold("Ship Count:\t") <> int.to_string(agent.ship_count))
}

pub fn server_status(
  server_status: ServerStatus,
  print_links: Bool,
  print_leaderboards: Bool,
) -> Nil {
  let now = timestamp.system_time()
  io.println(ansi.bold(ansi.underline(ansi.green("Server"))))
  io.println(ansi.bold("Version:\t") <> server_status.version)
  io.println(ansi.bold("Status:\t\t") <> server_status.status)
  io.println(
    ansi.bold("Last Reset:\t")
    <> legible_difference(
      now,
      timestamp.from_calendar(
        server_status.reset_date,
        TimeOfDay(0, 0, 0, 0),
        calendar.utc_offset,
      ),
    ),
  )
  io.println(
    ansi.bold("Next Reset:\t")
    <> legible_difference(now, server_status.server_resets.next)
    <> " ("
    <> ansi.italic(server_status.server_resets.frequency)
    <> ")",
  )
  case server_status.health.last_market_update {
    Some(last_market_update) ->
      io.println(
        ansi.bold("Market Updated:\t")
        <> legible_difference(now, last_market_update),
      )
    None -> Nil
  }
  io.println("")

  case print_links {
    True -> {
      io.println(ansi.bold(ansi.underline(ansi.blue("Links"))))
      list.each(server_status.links, fn(link) {
        io.println(ansi.bold(link.name) <> ": " <> uri.to_string(link.url))
      })
      io.println("")
    }
    False -> Nil
  }
  case print_leaderboards {
    True -> {
      io.println(
        ansi.bold(ansi.underline(ansi.yellow("Leaderboard - Most Credits"))),
      )
      list.each(server_status.leaderboards.most_credits, fn(entry) {
        io.println(
          ansi.bold(agent_symbol.to_string(entry.agent_symbol))
          <> ": "
          <> int.to_string(entry.credits),
        )
      })
      io.println("")

      io.println(
        ansi.bold(
          ansi.underline(ansi.yellow("Leaderboard - Most Submitted Charts")),
        ),
      )
      list.each(server_status.leaderboards.most_submitted_charts, fn(entry) {
        io.println(
          ansi.bold(agent_symbol.to_string(entry.agent_symbol))
          <> ": "
          <> int.to_string(entry.chart_count),
        )
      })
    }
    False -> Nil
  }
}

pub fn contract(contract: Contract) -> Nil {
  let now = timestamp.system_time()
  io.println(ansi.bold(ansi.underline(ansi.yellow("Contract"))))
  io.println(ansi.bold("Id:\t\t") <> contract_id.to_string(contract.id))
  io.println(
    ansi.bold("Status:\t\t")
    <> case contract.accepted {
      True ->
        case contract.fulfilled {
          True -> "FULFILLED"
          False -> "ACCEPTED"
        }
      False -> "NOT ACCEPTED"
    },
  )
  io.println(
    ansi.bold("Faction:\t") <> faction_symbol.to_string(contract.faction_symbol),
  )
  io.println(ansi.bold("Type:\t\t") <> contract_type.to_string(contract.type_))
  io.println(
    ansi.bold("Payment:\t")
    <> int.to_string(contract.terms.payment.on_accepted)
    <> " (on accept) + "
    <> int.to_string(contract.terms.payment.on_fulfilled)
    <> " (on fulfill) = "
    <> int.to_string(
      contract.terms.payment.on_accepted + contract.terms.payment.on_fulfilled,
    ),
  )
  io.println(
    ansi.bold("Deadlines:\t")
    <> case contract.deadline_to_accept {
      Some(deadline) ->
        "Accept by "
        <> legible_timestamp(deadline)
        <> " ("
        <> ansi.italic(legible_difference(now, deadline))
        <> ")"
      None -> "No deadline to accept"
    }
    <> " / Fulfill by "
    <> legible_timestamp(contract.terms.deadline)
    <> " ("
    <> ansi.italic(legible_difference(now, contract.terms.deadline))
    <> ")",
  )
  io.println(
    ansi.bold("Deliver:\t")
    <> case contract.terms.deliver {
      Some(goods) ->
        goods
        |> list.map(fn(good) {
          int.to_string(good.units_required)
          <> "x "
          <> trade_symbol.to_string(good.trade_symbol)
          <> " to "
          <> waypoint_symbol.to_string(good.destination_symbol)
        })
        |> string.join(", ")
      None -> "Nothing to deliver"
    },
  )
}

pub fn ship(ship: Ship) -> Nil {
  let now = timestamp.system_time()
  io.println(ansi.bold(ansi.underline(ansi.blue("Ship"))))
  io.println(ansi.bold("Symbol:\t\t") <> ship_symbol.to_string(ship.symbol))
  io.println(
    ansi.bold("Role:\t\t") <> ship_role.to_string(ship.registration.role),
  )
  io.println(
    ansi.bold("Faction:\t")
    <> faction_symbol.to_string(ship.registration.faction_symbol),
  )
  io.println(
    ansi.bold("Fuel:\t\t")
    <> int.to_string(ship.fuel.current)
    <> "/"
    <> int.to_string(ship.fuel.capacity)
    <> " ("
    <> ansi.italic(case ship.fuel.consumed {
      Some(ShipFuelConsumed(amount: 0, ..)) | None -> "none consumed"
      Some(consumed) ->
        "consumed "
        <> int.to_string(consumed.amount)
        <> " on "
        <> legible_timestamp(consumed.timestamp)
    })
    <> ")",
  )
  io.println(
    ansi.bold("Crew:\t\t")
    <> int.to_string(ship.crew.current)
    <> "/"
    <> int.to_string(ship.crew.capacity)
    <> " ("
    <> ansi.italic(int.to_string(ship.crew.required) <> " required")
    <> " / "
    <> ansi.italic(int.to_string(ship.crew.morale) <> " morale")
    <> " / "
    <> ansi.italic(crew_rotation.to_string(ship.crew.rotation) <> " rotation")
    <> ")",
  )
  io.println(
    ansi.bold("Cargo:\t\t")
    <> int.to_string(ship.cargo.units)
    <> "/"
    <> int.to_string(ship.cargo.capacity),
  )
  io.println(
    ansi.bold("Cooldown:\t")
    <> case ship.cooldown.expiration {
      Some(expiration) ->
        "Expires at "
        <> legible_timestamp(expiration)
        <> " ("
        <> ansi.italic(
          "in " <> int.to_string(ship.cooldown.remaining_seconds) <> " seconds",
        )
        <> ")"
      None -> ansi.italic("No active cooldown")
    },
  )
  io.println(
    ansi.bold("Location:\t")
    <> ship_nav_status.to_string(ship.nav.status)
    <> " ("
    <> ansi.italic(
      "in " <> ship_nav_flight_mode.to_string(ship.nav.flight_mode) <> " mode",
    )
    <> ")"
    <> " at "
    <> waypoint_symbol.to_string(ship.nav.waypoint_symbol),
  )
  io.println(
    ansi.bold("Origin:\t\t")
    <> case timestamp.compare(now, ship.nav.route.departure_time) {
      order.Lt -> "Departing"
      _ -> "Departed"
    }
    <> " "
    <> waypoint_type.to_string(ship.nav.route.origin.type_)
    <> " "
    <> waypoint_symbol.to_string(ship.nav.route.origin.symbol)
    <> " ("
    <> ansi.italic(
      int.to_string(ship.nav.route.origin.x)
      <> ", "
      <> int.to_string(ship.nav.route.origin.y),
    )
    <> ") at "
    <> legible_timestamp(ship.nav.route.departure_time)
    <> " ("
    <> ansi.italic(legible_difference(now, ship.nav.route.departure_time))
    <> ")",
  )
  io.println(
    ansi.bold("Destination:\t")
    <> case timestamp.compare(now, ship.nav.route.arrival) {
      order.Lt -> "Arriving"
      _ -> "Arrived"
    }
    <> " at "
    <> waypoint_type.to_string(ship.nav.route.destination.type_)
    <> " "
    <> waypoint_symbol.to_string(ship.nav.route.destination.symbol)
    <> " ("
    <> ansi.italic(
      int.to_string(ship.nav.route.destination.x)
      <> ", "
      <> int.to_string(ship.nav.route.destination.y),
    )
    <> ") at "
    <> legible_timestamp(ship.nav.route.arrival)
    <> " ("
    <> ansi.italic(legible_difference(now, ship.nav.route.arrival))
    <> ")",
  )
}
